#ifdef WIN32
#include <windows.h>
#endif


#include <GL/glew.h>
#include <GL/glut.h>

#include <IL/il.h>
#include <IL/ilut.h>

#include <cstdlib>
#include <algorithm>

#include "glutil.h"
#include "float4x4.h"
#include "float3x3.h"

#include <OBJModel.h>


using std::min;
using std::max;
using namespace chag;

// Used for standard shading
GLuint shaderProgram;
// Used for drawing light and drawing to shadowMap
GLuint simpleShaderProgram; 


float currentTime = 0.0f;

bool trigSpecialEvent = false;
bool paused = false;

// Scene up orientation, used when creating view matrices, handy to define somewhere...
const float3 up = {0.0f, 1.0f, 0.0f};

float camera_theta = M_PI / 4.0f; 
float camera_phi = M_PI / 6.0f; 
float camera_r = 30.0; 

float3 sphericalToCartesian(float theta, float phi, float r)
{
	return make_vector( r * sinf(theta)*sinf(phi),
					 	r * cosf(phi), 
						r * cosf(theta)*sinf(phi) );
}


// lights angle around y-axis
float3 lightPosition; 

// Model matrices
float4x4 roomModelMatrix;
float4x4 fighterModelMatrix;


OBJModel *fighterModel = 0;
OBJModel *boxModel = 0;


void initGL()
{
	// Initialize GLEW, which provides access to OpenGL Extensions
	glewInit();  

	// Print infomation about GL and ensure that we've got GL.30
	startupGLDiagnostics();

	ilInit();					// initialize devIL (developers Image Library)
	ilutRenderer(ILUT_OPENGL);  // initialize devIL

	// Workaround for AMD, which hopefully will not be neccessary in the near future...
	if (!glBindFragDataLocation)
	{
		glBindFragDataLocation = glBindFragDataLocationEXT;
	}

	//************************************
	//		Load Shaders
	//************************************
	shaderProgram = loadShaderProgram("shading.vert", "shading.frag");
	  glBindAttribLocation(shaderProgram, 0, "position"); 	
	  glBindAttribLocation(shaderProgram, 1, "normalIn");
	  glBindAttribLocation(shaderProgram, 2, "texCoordIn");
	  glBindFragDataLocation(shaderProgram, 0, "fragmentColor");
	linkShaderProgram(shaderProgram);

	simpleShaderProgram = loadShaderProgram("simple.vert", "simple.frag");
	  glBindAttribLocation(simpleShaderProgram, 0, "position"); 	
	  glBindFragDataLocation(simpleShaderProgram, 0, "fragmentColor");
	linkShaderProgram(simpleShaderProgram);

	CHECK_GL_ERROR();

	//************************************
	// Load models and set up model matrices
	//************************************

	fighterModel = new OBJModel;
	fighterModel->load("../scenes/fighter.obj");

  boxModel = new OBJModel;
  boxModel->load("../scenes/lab5.obj");
  
  // Note:
  roomModelMatrix = make_scale<float4x4>(make_vector(3.5f, 1.0f, 3.5f));

	fighterModelMatrix = make_translation(make_vector(0.0f, 3.5f, 0.0f))
                          * make_rotation_y<float4x4>(currentTime * -M_PI / 4.0f)
                          * make_scale<float4x4>(0.2f);



	glEnable(GL_DEPTH_TEST);	// enable Z-buffering 
	glEnable(GL_CULL_FACE);		// enables backface culling
}


/**
 * Helper function to set the matrices used for transform an lighting in our shaders,
 * this code is factored into its own function as we use it more than once (needed
 * once for each model drawn). This is generally good software engineering practice.
 */
void setLightingMatrices(GLuint shaderProgram, const float4x4 &viewMatrix, const float4x4 &projectionMatrix, const float4x4 &modelMatrix)
{
	float4x4 modelViewMatrix = viewMatrix * modelMatrix;	
	float4x4 modelViewProjectionMatrix = projectionMatrix * modelViewMatrix;
	float4x4 normalMatrix = transpose(inverse(modelViewMatrix));
	// Update the matrices used in the vertex shader
	setUniformSlow(shaderProgram, "modelViewMatrix", modelViewMatrix);
	setUniformSlow(shaderProgram, "modelViewProjectionMatrix", modelViewProjectionMatrix);
	setUniformSlow(shaderProgram, "normalMatrix", normalMatrix);
}


/**
 * In this function, we draw all scene elements that should cast shadow. 
 * This function is called twice, once for rendering the shadow map, and once when drawing the visible scene.
 * The code could have been duplicated, but again, it makes the code much easier to manage, for example 
 * if we were to add more objects to the scene.
 */
void drawShadowCasters(GLuint shaderProgram, const float4x4 &viewMatrix, const float4x4 &projectionMatrix)
{
	/**************************
	 * Draw the room
	 *************************/
  setLightingMatrices(shaderProgram, viewMatrix, projectionMatrix, roomModelMatrix);
  boxModel->render();
	/**************************
	 * Draw the space ship
	 *************************/
  setLightingMatrices(shaderProgram, viewMatrix, projectionMatrix, fighterModelMatrix);
  fighterModel->render();
}



void drawScene(const float4x4 &viewMatrix, const float4x4 &projectionMatrix,
					  const float4x4 &lightViewMatrix, const float4x4 &lightProjectionMatrix)
{
	int w = glutGet((GLenum)GLUT_WINDOW_WIDTH);
	int h = glutGet((GLenum)GLUT_WINDOW_HEIGHT);

  // Set viewport
	glViewport(0, 0, w, h);								

  // Set clear color
	glClearColor(0.2,0.2,0.8,1.0);						
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); // Clears the color buffer and the z-buffer

	// Draw light for reference
  debugDrawLight(viewMatrix, projectionMatrix, lightPosition); 

  // Shader Program
	glUseProgram( shaderProgram );				// Set the shader program to use for this draw call

  // set the 0th texture unit to serve the 'diffuse_texture' sampler.
	setUniformSlow(shaderProgram, "diffuse_texture", 0 );
	glActiveTexture(GL_TEXTURE0); 

	// Set the lights view space coordinates to the shaders
	float3 viewSpaceLightPos = transformPoint(viewMatrix, lightPosition); 
	setUniformSlow(shaderProgram, "viewSpaceLightPosition", viewSpaceLightPos);


	CHECK_GL_ERROR(); 


  drawShadowCasters(shaderProgram, viewMatrix, projectionMatrix);
	
	glUseProgram( 0 );	
	CHECK_GL_ERROR();
}



void drawShadowMap(const float4x4 &viewMatrix, const float4x4 &projectionMatrix)
{
	glClearColor(1.0, 1.0, 1.0, 1.0);
	glClearDepth(1.0);
	glClear(GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT);
	// Shader Program
	GLint current_program; 
	glGetIntegerv(GL_CURRENT_PROGRAM, &current_program);
	glUseProgram( simpleShaderProgram );


  drawShadowCasters(simpleShaderProgram, viewMatrix, projectionMatrix);

	glUseProgram( current_program );	

	CHECK_GL_ERROR();
}



void display(void)
{
	CHECK_GL_ERROR();

	// Now set up view and projection matrices for light
	float4x4 lightViewMatrix = lookAt(lightPosition, make_vector(0.0f, 0.0f, 0.0f), up);
	float4x4 lightProjectionMatrix = perspectiveMatrix(45.0f, 1.0, 5.0f, 100.0f);


	int w = glutGet((GLenum)GLUT_WINDOW_WIDTH);
	int h = glutGet((GLenum)GLUT_WINDOW_HEIGHT);

	float4x4 viewMatrix = lookAt(sphericalToCartesian(camera_theta, camera_phi, camera_r), 
								               make_vector(0.0f, 0.0f, 0.0f),	
								               up);	
	float4x4 projectionMatrix = perspectiveMatrix(45.0f, float(w) / float(h), 0.01f, 300.0f);

	// we then pass these matrices to the scene drawing function.
	drawScene(viewMatrix, projectionMatrix, lightViewMatrix, lightProjectionMatrix);

	CHECK_GL_ERROR();

	glutSwapBuffers();  // swap front and back buffer. This frame will now be displayed.
	CHECK_GL_ERROR();
}



void handleKeys(unsigned char key, int /*x*/, int /*y*/)
{
	switch(key)
	{
	case 27:    /* ESC */
		exit(0); /* dirty exit */
		break;   /* unnecessary, I know */
	case 32:    /* space */
		break;
	}
}



void handleSpecialKeys(int key, int /*x*/, int /*y*/)
{
	switch(key)
	{
	case GLUT_KEY_LEFT:
		printf("Left arrow\n");
		break;
	case GLUT_KEY_RIGHT:
		printf("Right arrow\n");
		break;
	case GLUT_KEY_UP:
	case GLUT_KEY_DOWN:
		break;
	}
}



bool leftDown = false;
bool middleDown = false;
bool rightDown = false;

int prev_x = 0;
int prev_y = 0;



void mouse(int button, int state, int x, int y)
{
	// reset the previous position, such that we only get movement performed after the button
	// was pressed.
	prev_x = x;
	prev_y = y;

	bool buttonDown = state == GLUT_DOWN;

	switch(button)
	{
	case GLUT_LEFT_BUTTON:
		if(leftDown != buttonDown)
			trigSpecialEvent = !trigSpecialEvent;
		leftDown = buttonDown;
		break;
	case GLUT_MIDDLE_BUTTON:
		middleDown = buttonDown;
		break;
	case GLUT_RIGHT_BUTTON: 
		rightDown = buttonDown;
	default:
		break;
	}
}



void motion(int x, int y)
{
	int delta_x = x - prev_x;
	int delta_y = y - prev_y;
	
	if(middleDown)
	{
		camera_r -= float(delta_y) * 0.3f;
		// make sure cameraDistance does not become too small
		camera_r = max(0.1f, camera_r);
	}
	if(leftDown)
	{
		camera_phi	-= float(delta_y) * 0.3f * float(M_PI) / 180.0f;
		camera_phi = min(max(0.01f, camera_phi), float(M_PI) - 0.01f);
		camera_theta += float(delta_x) * 0.3f * float(M_PI) / 180.0f;
	}
	prev_x = x;
	prev_y = y;
}



void idle( void )
{
	// glutGet(GLUT_ELAPSED_TIME) returns the time since application start in milliseconds.

	// this is updated the first time we enter this function, otherwise we will take the
	// time from the start of the application, which can sometimes be long.
	static float startTime = float(glutGet(GLUT_ELAPSED_TIME)) / 1000.0f;
	// update the global time if the application is not paused.
	if (!paused)
	{
		currentTime = float(glutGet(GLUT_ELAPSED_TIME)) / 1000.0f - startTime;
	}

	// Here is a good place to put application logic.
	float lightYAngle = 0.5f * currentTime; 
	lightPosition = make_rotation_y<float3x3>(lightYAngle) * make_vector(10.0f, 10.0f, 0.0f); 

	glutPostRedisplay(); 
}



int main(int argc, char *argv[])
{
#	if defined(__linux__)
	linux_initialize_cwd();
#	endif // ! __linux__

	glutInit(&argc, argv);
	/* open a window with double buffering, RGB colors, and Z-buffering */
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
	glutInitWindowSize(512,512);
	glutCreateWindow("OpenGL Lab 6");
	glutKeyboardFunc(handleKeys);
	glutSpecialFunc(handleSpecialKeys);
	/* the display function is called once when the glutMainLoop is called,
	* but also each time the window has to be redrawn due to window 
	* changes (overlap, resize, etc). It can also be forced to be called
	* by making a glutPostRedisplay() call 
	*/
	glutDisplayFunc(display);	// This is the main redraw function
	glutMouseFunc(mouse);		// callback function on mouse buttons
	glutMotionFunc(motion);		// callback function on mouse movements
	glutIdleFunc(idle);

	glutDisplayFunc(display);	// Set the main redraw function

	initGL();

	glutMainLoop();  /* start the program main loop */

	return 0;          
}
